
import sympy as sp
from itertools import combinations

from sympy import ImmutableDenseMatrix as IMat
from sympy import Matrix as Mat
from sympy import Rational as Rational


def get_test_points():
    points=[IMat([3,0,0]),IMat([2,2,0]),IMat([-2,2,0]),IMat([Rational(3,2),Rational(-3,2),2]),IMat([Rational(3,2),Rational(3,2),2]),IMat([Rational(-3,2),Rational(-3,2),2]),IMat([Rational(-3,2),Rational(3,2),2]),IMat([0,0,3])]
    all_points=[]
    for p in points:
        all_points.append(p)
        all_points.append(-p)
    return all_points

def get_test_face():
    p1 = Vertex(IMat([0,0,3]))
    p2 = Vertex(IMat([0,0,-3]))
    p3 = Vertex(IMat([3,0,0]))
    p4 = Vertex(IMat([-3,0,0]))
    e1 = Polyhedron(1,[p1,p3])
    e2 = Polyhedron(1,[p3,p2])
    e3 = Polyhedron(1,[p2,p4])
    e4 = Polyhedron(1,[p4,p1])
    f = Polyhedron(2,[e1,e2,e3,e4])
    return f
               

class Polyhedron():
    def __init__(self,dim,facets=[]):
        self.dim = dim
        self._facets = {dim-1:facets}
        if self.dim > 1:
            for d in range(1,dim-1):
                self._facets[d]=[]
                for F in self._facets[dim-1]:
                    for f in F.facets(d):
                        if [v.coords() for v in f.vertices()] not in [[v.coords() for v in self._facets[d][i].vertices()] for i in range(len(self._facets[d]))]:
                            self._facets[d].append(f)
            verts=[]
            for F in self._facets[dim-1]:
                for v in F.vertices():
                    if v.coords() not in [w.coords() for w in verts]:
                        verts.append(v)
            self._facets[0] = tuple(verts)

    def __repr__(self):
        verts = ''
        for v in self.vertices():
            verts=verts+str(tuple(v.coords()[i] for i in range(len(v.coords()))))+','

        return "Convex hull of vertices: "+verts[:-1]

    def __hash__(self):
        return hash(self.vertices())

    def __eq__(self,other):
        return ( self.__class__ == other.__class__ ) and ( self.vertices() == other.vertices() )

    def __ne__(self,other):
        return ( self.__class__ != other.__class__ ) or ( self.vertices() != other.vertices() )

    def facets(self,d):
            return self._facets[d]

    def vertices(self):
            return self._facets[0]

    def is_subfacet(self,f1,f2):
        if f1 in f2.facets(f1.dim):
                return True
        else:
            return False

    def neighbors(self,vertex):
        neighbors = []
        for e in self.facets(1):
            if self.is_subfacet(vertex,e):
                for v in e.vertices():
                    if v != vertex:
                        neighbors.append(v)
        return neighbors


class Vertex():
    def __init__(self,coords):
        self._coords = coords
        self.dim = 0

    def __repr__(self):
        return str(tuple(self._coords[i] for i in range(len(self._coords))))

    def __hash__(self):
        return hash(self._coords)

    def __eq__(self,other):
        return ( self.__class__ == other.__class__ ) and self._coords == other._coords

    def __ne__(self,other):
        return ( self.__class__ != other.__class__ ) or self._coords != other._coords

    def coords(self,dim=0):
        return self._coords



def ns(vector):   ## norm squared
    return vector.norm()**2



def cone_on_face(face,vertex):
    if face.dim == 0:
        return Polyhedron(1,[face,vertex])
    else:
        facets = [face]
        for f in face.facets(face.dim-1):
            facets.append(cone_on_face(f,vertex))
        return Polyhedron(face.dim+1,facets)

def quick_hull(points, dim):
    # choose dim+1 points for the initial hull. We might as well pick the ones of largest norm.
    v_max = sorted([(ns(v),v) for v in points],key=lambda x:x[0])[-1][1]
    B_points = [v_max]
    while len(B_points)<dim:
        A = Mat([v.transpose() for v in B_points]).transpose()
        Proj_P = A*(A.transpose()*A).inv()*A.transpose() #projection to the subspace generated by B_points
        perp_norms = dict([(v,ns(v-Proj_P*v)) for v in points])
        max_perp = sorted([(perp_norms[v],v) for v in points],key=lambda x:x[0])[-1][1]
        B_points.append(max_perp)
    #O = Vertex(IMat([0 for i in range(dim)]))
    F = simplex(B_points,dim-1)
    facets = []
    for set1,set2 in partitions(F.vertices()):
        facets.append(simplex([-v.coords() for v in set1]+[v.coords() for v in set2],dim-1))
        facets.append(simplex([-v.coords() for v in set2]+[v.coords() for v in set1],dim-1))
    hull = Polyhedron(dim,facets)
    above = {} # this dictionary will associate with each face the vertices that are above it 
                 # (i.e., those having positive inner product with the outward normal of the face)
    for F in hull.facets(dim-1):
        above[F] = outside_of(F,points) 
    points = set([p for F in hull.facets(hull.dim-1) for p in above[F]]) # we can get rid of points not above at least one face
    while len(points) > 0: # as long as there are still points outside of the hull, expand faces to points that lie above them
        for F in hull.facets(dim-1):
            if len(above[F]) > 0:
                p = above[F][-1] # pick the point furthest from F (i.e., the last in the list)
        hull,points,above = expand(hull,p,points,above)

    #### ***** still need to combine co-planar faces. Compute a normal for each face, group by scalar multiple, combine faces in group.********

    return hull

def expand(hull,point,points,above):
    visible = [face for face in hull.facets(hull.dim-1) if point in above[face]] # faces visible from point
    horizons = [] # facets of dimension dim-2 that are on the boundary of the union of visible faces as seen from point
    for face in visible:
        for subface in face.facets(hull.dim-2):
            on_faces = [f for f in visible if hull.is_subfacet(subface,f)]
            if len(on_faces) == 1: # only add to horizons if it is not shared by multiple faces (so it's not in the interior)
                horizons.append(subface)
    faces = hull.facets(hull.dim-1) # all faces of current hull
    for face in visible:
        faces.remove(face) #remove visible faces as they are now covered by the new faces we'll create below
        del above[face] # also remove these from the above dictionary
    points.discard(point)
    for h in horizons:
        new_face = cone_on_face(h,Vertex(point)) # new faces come from coning over the horizons to point
        above[new_face] = outside_of(new_face,points) # populate above[new_face] with points that lie above new_face
        faces.append(new_face) # add to faces each new face

    hull = Polyhedron(hull.dim,faces) # construct the new hull
    points = set([p for F in hull.facets(hull.dim-1) for p in above[F]])
    return hull, points, above


def simplex(points,dim):
    assert len(points) == dim+1
    if dim == 1:
        P = Polyhedron(1,[Vertex(points[0]),Vertex(points[1])])
        return P
    else:
        facets = [simplex(f,dim-1) for f in list(combinations(points,dim))]
        P = Polyhedron(dim,facets)
        return P


def outside_of(face,points):
    outside = []
    basepoint = face.vertices()[0]  # pick a vertex to translate everything to the origin by
    pre_basis = face.neighbors(basepoint)
    dim = face.dim
    basis = [v.coords()-basepoint.coords() for v in pre_basis][:dim]
    A = Mat([v.transpose() for v in basis]).transpose()
    Proj_P = A*((A.transpose()*A).inv())*A.transpose()
    #C = Mat([0 for i in range(len(basis))])
    #A_t=A.transpose()

    #solutions, param = A_t.gauss_jordan_solve(C)  # solve the system of equations, returns parametrization of solutions
    #param_one = { tau:1 for tau in param } #set parameters to 0 to get a solution which will hopefully be an integer solution.
    #normal = solutions.xreplace(param_one)
    #if normal.dot(basepoint.coords()) < 0:
    #    normal = -normal
    normal = basepoint.coords() - Proj_P*basepoint.coords()
    for p in points:
        dot = normal.dot(p-basepoint.coords())
        if dot > 0:
            outside.append((dot,p))
    outside.sort(key=lambda x: x[0])
    return [pair[1] for pair in outside]  


def partitions(collection):
    collection = list(collection)
    if len(collection) == 1:
        return [[ collection,[] ]]

    parts = []
    first = collection[0]
    for [set1, set2] in partitions(collection[1:]):
        # insert `first` in each of the subpartition's subsets
        parts.append([set1,set2+[first]])
        parts.append([set1+[first],set2])
    return parts











import snappy

from tnorm.kernel import matrices
from tnorm.utilities.regina_helpers import regina_to_sage_int, regina_to_sage_mat
from tnorm.utilities.sage_types import *



def get_quads(spun_surface,tet,v1,v2):
    s = spun_surface
    tet = int(tet)
    v2 = int(v2)
    return regina_to_sage_int(s.quads(tet,v2-int(1)))


### this maps an oriented spun normal surface to H_1(bdy M; \ZZ)
def signed_bdy_maps(oriented_spun_surface, peripheral_curve_mats, spinning=False, oriented_quads_mat=None):
    s = oriented_spun_surface
    T = s.triangulation()
    quads_mat = oriented_quads_mat


    pos_intx_mat = matrices.pos_intersection_mat()
    neg_intx_mat = matrices.neg_intersection_mat()
    if spinning:
        neg_intx_mat = neg_intx_mat.apply_map(abs)
    pos_slopes = []                   # on a surface affects boundary orientations.
    neg_slopes = []
    if quads_mat == None:
        quads_mat = matrices.oriented_quads_mat(s)

    for c in range(T.countCusps()):
        periph_mat = peripheral_curve_mats[c] # get the matrix that encodes meridian and longitude for cusp c.
        
        iota_lambda_pos=0
        iota_mu_pos=0
        iota_lambda_neg=0
        iota_mu_neg=0
        for i in range(T.size()):
            iota_lambda_pos += Integer(periph_mat[0][i]*pos_intx_mat*quads_mat[i])
            iota_mu_pos += Integer(periph_mat[1][i]*pos_intx_mat*quads_mat[i])
            iota_lambda_neg += Integer(periph_mat[0][i]*neg_intx_mat*quads_mat[i])
            iota_mu_neg += Integer(periph_mat[1][i]*neg_intx_mat*quads_mat[i])

  
        pos_slopes.append((-iota_lambda_pos,iota_mu_pos))
        neg_slopes.append((-iota_lambda_neg,iota_mu_neg))
    return pos_slopes, neg_slopes

### this is almost the same as sign_bdy_matrix(), except we take the (entry-wise) absolute value
### of neg_intersection_mat(). This puts the orientation on quad edges described in Tillmann-- "Normal 
### surfaces in topologically finite 3-manifolds". The result are boundary slopes with signs indicating
### spinning direction
def spinning_bdy_maps(oriented_spun_surface, peripheral_curve_mats):
    return signed_bdy_maps(oriented_spun_surface, peripheral_curve_mats, True)

## this map is almost the same as qtons_to_H1bdy(), except the SIGN_MATRIX is replaced by 
## SIGN_MATRIX.apply_map(abs). The result is the boundary slope for the embedded, possibly non-orientable,
## normal surface obtained by forgetting orientation. The reason this could be non-orientable is 
## that boundaries of the same slope that spin in oposite directions are cancelled. Thus the
## self-intersections coming from spinning are avoided, but at the cost of cutting and gluing
## in a way that does not respect orientation. This should return slopes which are (up to sign)
## the same as those returned by s.boundaryIntersections().
def bdy_slopes_unoriented_(oriented_spun_surface, peripheral_curve_mats, quad_mat=None):
    s = oriented_spun_surface
    T = s.triangulation()


    intersection_mat = matrices.intersection_mat().apply_map(abs)
    slopes = [] 
    for c in range(T.countCusps()):
        periph_mat = peripheral_curve_mats[c] # get the matrix that encodes meridian and longitude for cusp c.
        if quad_mat == None:
            quad_mat = matrices.oriented_quads_mat(s) # get the matrix that encodes quad coordinates for s.
        iota_lambda = QQ(sum([periph_mat[0][i]*intersection_mat*quad_mat[i] for i in range(T.size())]))
        iota_mu = QQ(sum([periph_mat[1][i]*intersection_mat*quad_mat[i] for i in range(T.size())]))
        slopes.append((-iota_lambda,iota_mu))
    return slopes











import sys
import argparse
import tnorm.GUI.application

start = tnorm.GUI.application.start



if __name__ == '__main__':
#    parser = argparse.ArgumentParser(description='tnorm GUI')
#    parser.add_argument('load', nargs='?', help='path to load from when starting')
#    args = parser.parse_args()
    
#    start(load_from=args.load)
    start()




from tnorm.kernel import embedded, orientable, connected, euler
from tnorm.kernel.matrices import peripheral_curve_mats
from tnorm.kernel.boundary import signed_bdy_maps
import snappy


def orient(S):
    return orientable.orient(S)

def is_orientable(S):
    return orientable.is_orientable(S)

def euler_char(S):
    return euler.euler_char_(S,euler.solve_lin_gluing_eq(S.triangulation()))

def oriented_boundary_slopes(S):
    T = S.triangulation()
    M = snappy.Manifold(T.snapPea())
    pc_mats = peripheral_curve_mats(M,T)
    pos_bdy, neg_bdy = signed_bdy_maps(S, pc_mats, False, orient(S))
    return {'outward':pos_bdy, 'inward':neg_bdy}

def connected(S):
    pass


from tnorm.utilities.regina_helpers import *
from tnorm.kernel.matrices import oriented_quads_mat, quads_mat

from tnorm.kernel.orientable import orient

def ends_embedded(qtons, TN_wrapper):
    W = TN_wrapper
    for i in range(W.num_cusps()):
        pos_bdy = W.spinning_slopes(qtons)[0][i]
        neg_bdy = W.spinning_slopes(qtons)[1][i]
        gcd_pos = gcd(pos_bdy[0],pos_bdy[1])
        gcd_neg = gcd(neg_bdy[0],neg_bdy[1])
        if gcd_pos != 0 and gcd_neg != 0:
            if not ( pos_bdy[0]/gcd_pos == neg_bdy[0]/gcd_neg and pos_bdy[1]/gcd_pos == neg_bdy[1]/gcd_neg ):
                return False
    return True


def is_embedded(qtons,TN_wrapper):
    W = TN_wrapper
    S = qtons
    T = W.triangulation()

    if not W.manifold_is_closed():
        pos_bdy, neg_bdy = W.boundary_slopes(S).values()
        if not ends_embedded(S,W):
            return False

    oriented_projection = orient(S)
    
    if oriented_projection != False:
        opp_oriented_projection = []
        for row in oriented_projection:
            opp_row = []
            for i in [0,2,4]:
                opp_row.append(row[i+1])
                opp_row.append(row[i])
            opp_oriented_projection.append(opp_row)

        q_mat = oriented_quads_mat(S)
        
        if oriented_projection == q_mat or Matrix(opp_oriented_projection) == q_mat:
            return True
    return False


def gcd(a,b):
    return abs(GCD(a,b))

















from tnorm.kernel.abstract_nbhd import abstract_nbhds





def is_connected(spun_normal_surface):
    S = spun_normal_surface
    T = S.triangulation()

    abs_nbhds = abstract_nbhds(S)

    comp_indices = [(k,j) for k in abs_nbhds for j in range(abs_nbhds[k]['components'].dimensions()[0])]

    stacks = {i:[] for i in range(T.size())}
    for j in abs_nbhds:
        comps = abs_nbhds[j]['components']
        for k in range(comps.dimensions()[1]):
            tet = abs_nbhds[j]['tets'][k]
            stacks[tet].append([(j,k),comps.column(k),abs_nbhds[j]['align'][k]])
            
    #return comp_indices, stacks



import snappy
import tnorm.kernel.simplicial
import regina


has_internal_homology = ['s789', 'v1539', 'v3209', 't06828', 't07788', 't08868', 't09692', 't12061','o9_09412', 'o9_23005', 'o9_23006', 'o9_31901', 'o9_32877', 'o9_36416', 'o9_36417', 'o9_36580', 'o9_40732', 'o9_41691', 'o9_41693', 'o9_44004', 'o9_44007', 'o9_44157']

closed_isosigs = ['kLLzvQQkacegjihihjjnkxmeqxjjcn', 'kLvPPPAkafedhhiigjjjqxnqxahnhj', 'kLvzQQzkccfehghgijjhnhqckcnahn', 'kLvzQQzkccfehghgijjhrhqckcnahr', 'kLLLAPPkcdgfehijijjhshaqpihkkk', 'kLLwvQQkacdgijhjiijjkalilmckwo', 'kLLzLPQkacegjihihjjjkxpuqxrwvv', 'kLLzLPQkacegjihihjjnkxpuqxrwvv', 'kLLzLPQkaceihjgijjijkxuetujjgs', 'kLLzLPQkaceihjgijjinkxuetujjgs']

links = ['K10a110', 'L14n43208', 'L12n2026', 'K10n22', 'L12n1054', 'K10n13', 'L14n49678', 'L7a6', 'L8a18']

tests = [closed_isosigs,has_internal_homology,links]

def test(do_tests=[0,1,2]):

    if 0 in do_tests:
        for name in closed_isosigs:
            print(name)
            T=regina.Triangulation3.fromIsoSig(name)
            W = tnorm.load(T, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
#            assert B.polyhedron.is_compact(), W.manifold.name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'
        
            W = tnorm.load(name, quiet=True, force_simplicial_homology=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=True'
            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=True'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=True'
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=True'
#            assert B.polyhedron.is_compact(), W.manifold.name()+', force_simplicial=True'

    
    if 1 in do_tests:
        for name in has_internal_homology:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=False'
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=True'
            
            W = tnorm.load(name, quiet=True, force_simplicial_homology=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=True'
            assert W.has_internal_homology(), W.manifold().name()+', force_simplicial=True'
            del3 = tnorm.kernel.simplicial.del3_matrix(W.triangulation())
            del2 = tnorm.kernel.simplicial.del2_matrix(W.triangulation())
            assert (del2*del3).is_zero(), W.manifold().name()+', force_simplicial=True'
            for i in range(W.qtons().size()):
                assert (del2*W.simplicial_class(i)).is_zero(), W.manifold().name()+', force_simplicial=True'
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=True'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=True'

    if 2 in do_tests:
        for name in links:
            print(name)
            W = tnorm.load(name, quiet=True)
            B = W.norm_ball
            assert B.polyhedron().dim() == W.triangulation().homologyH1().rank(), W.manifold().name()+', force_simplicial=False'
            assert not W.has_internal_homology(), W.manifold().name()+', force_simplicial=False'
            for i in range(W.qtons().size()):
                assert W.euler_char(i) == 2-2*W.genus(i)-W.num_boundary_comps(i), W.manifold().name()+', force_simplicial=False'
            assert B.polyhedron().is_compact(), W.manifold().name()+', force_simplicial=False'
            for i in range(len(B.vertices())):
                if B.vertices()[i].qtons_index != None:
                    assert W.is_embedded(B.vertices()[i].qtons_index()) == True, W.manifold().name()+', force_simplicial=False'
        
            if W.manifold().num_cusps()>1:
                W_sim = tnorm.load(name, quiet=True, force_simplicial_homology=True)
                B_sim = W_sim.norm_ball
                assert B_sim.polyhedron().dim() == W_sim.triangulation().homologyH1().rank(), W_sim.manifold().name()+', force_simplicial=True'
                assert not W_sim.has_internal_homology(), W_sim.manifold().name()+', force_simplicial=True'
                del3 = tnorm.kernel.simplicial.del3_matrix(W_sim.triangulation())
                del2 = tnorm.kernel.simplicial.del2_matrix(W_sim.triangulation())
                assert (del2*del3).is_zero(), W_sim.manifold().name()+', force_simplicial=True'
                for i in range(W_sim.qtons().size()):
                    assert (del2*W_sim.simplicial_class(i)).is_zero(), W_sim.manifold().name()+', force_simplicial=True'
                    assert W_sim.euler_char(i) == 2-2*W_sim.genus(i)-W_sim.num_boundary_comps(i), W_sim.manifold().name()+', force_simplicial=True'
                assert B_sim.polyhedron().is_compact(), W_sim.manifold().name()+', force_simplicial=True'
                P = B.polyhedron()

                





import sympy as sp
from itertools import combinations

from sympy import ImmutableDenseMatrix as IMat
from sympy import Matrix as Mat
from sympy import Rational as Rational
from sympy import QQ


def get_test_points():
    points=[IMat([QQ(3),QQ(0),QQ(0)]),IMat([QQ(2),QQ(2),QQ(0)]),IMat([QQ(-2),QQ(2),QQ(0)]),IMat([QQ(3,2),QQ(-3,2),QQ(2)]),IMat([QQ(3,2),QQ(3,2),QQ(2)]),IMat([QQ(-3,2),QQ(-3,2),QQ(2)]),IMat([QQ(-3,2),QQ(3,2),QQ(2)]),IMat([QQ(0),QQ(0),QQ(3)])]
    all_points=[]
    for p in points:
        all_points.append(p)
        all_points.append(-p)
    return all_points

def get_test_face():
    p1 = Vertex(IMat([0,0,3]))
    p2 = Vertex(IMat([0,0,-3]))
    p3 = Vertex(IMat([3,0,0]))
    p4 = Vertex(IMat([-3,0,0]))
    e1 = Polyhedron(1,[p1,p3])
    e2 = Polyhedron(1,[p3,p2])
    e3 = Polyhedron(1,[p2,p4])
    e4 = Polyhedron(1,[p4,p1])
    f = Polyhedron(2,[e1,e2,e3,e4])
    return f

Finally, we have l A M = d I, so (l/d A) M = I and l/d A is the inverse
you seek.A N = d I

### finds the convex hull P of a set of points in d dimensions. Assumes the convex hull is a centrally symmetric polyhedron.
### 




def hull(points, dim):
    ## first we find a hyperplane B such that B cuts P in half along a maximal area cross-section.
    ## this allows us to just compute the convex hull of points to one side of B, then reflect.
    v_max = sorted([(ns(v),v) for v in points],key=lambda x:x[0])[-1][1]
    B_points = [v_max]
    #points.remove(v_max)
    #points.remove(-v_max)
    while len(B_points)<dim-1:
        A = Mat([v.transpose() for v in B_points]).transpose()
        Proj_P = A*(A.transpose()*A).inv()*A.transpose() #projection to the subspace generated by B_points
        perp_norms = dict([(v,ns(v-Proj_P*v)) for v in points])
        max_perp = sorted([(perp_norms[v],v) for v in points],key=lambda x:x[0])[-1][1]
        B_points.append(max_perp)
        #points.remove(max_perp)
        #points.remove(-max_perp)
    # remove all points to one side of B

    A = Mat([v.transpose() for v in B_points]).transpose()
    Proj_P = A*(A.transpose()*A).inv()*A.transpose()
    C = Mat([0 for i in range(dim-1)])
    A_t=A.transpose()

    solutions, param = A_t.gauss_jordan_solve(C)  # solve the system of equations, returns parametrization of solutions
    param_one = { tau:1 for tau in param } #set parameters to 0 to get a solution which will hopefully be an integer solution.
    normal = solutions.xreplace(param_one)
    to_remove = []
    for v in points:
        if normal.dot(v) < 0:
            to_remove.append(v)
    for v in to_remove:
        points.remove(v)
    # now we create our initial hull which includes points of B_points and their negatives, and the furthest point from the plane B.
    perp_norms = dict([(v,ns(v-Proj_P*v)) for v in points])
    max_perp = sorted([(perp_norms[v],v) for v in points],key=lambda x:x[0])[-1][1]
    B_points = [v for v in B_points]+[-v for v in B_points]
    F = hull(B_points,dim-1) # take the convex hull of points in B_points to get an initial face
    hull = double_cone_on_face(F,max_perp) # get the initial hull by taking the cone over F to max_perp
    outside = {}
    for f in hull.facets(dim-1):
        outside[f] = outside_of(f,points)

    return hull_points,points

class Polyhedron():
    def __init__(self,dim,facets=[]):
        self.dim = dim
        self._facets = {dim-1:facets}
        if self.dim > 1:
            for d in range(1,dim-1):
                self._facets[d]=[]
                for F in self._facets[dim-1]:
                    for f in F.facets(d):
                        if [v.coords() for v in f.vertices()] not in [[v.coords() for v in self._facets[d][i].vertices()] for i in range(len(self._facets[d]))]:
                            self._facets[d].append(f)
            verts=[]
            for F in self._facets[dim-1]:
                for v in F.vertices():
                    if v.coords() not in [w.coords() for w in verts]:
                        verts.append(v)
            self._facets[0] = tuple(verts)

    def __repr__(self):
        verts = ''
        for v in self.vertices():
            verts=verts+str(tuple(v.coords()[i] for i in range(len(v.coords()))))+','

        return "Convex hull of vertices: "+verts[:-1]

    def __hash__(self):
        return hash(self.vertices())

    def __eq__(self,other):
        return ( self.__class__ == other.__class__ ) and ( self.vertices() == other.vertices() )

    def __ne__(self,other):
        return ( self.__class__ != other.__class__ ) or ( self.vertices() != other.vertices() )

    def facets(self,d):
            return self._facets[d]

    def vertices(self):
            return self._facets[0]

    def is_subfacet(self,f1,f2):
        if f1 in f2.facets(f1.dim):
                return True
        else:
            return False

    def neighbors(self,vertex):
        neighbors = []
        for e in self.facets(1):
            if self.is_subfacet(vertex,e):
                for v in e.vertices():
                    if v != vertex:
                        neighbors.append(v)
        return neighbors


class Vertex():
    def __init__(self,coords):
        self._coords = coords
        self.dim = 0

    def __repr__(self):
        return str(tuple(self._coords[i] for i in range(len(self._coords))))

    def __hash__(self):
        return hash(self._coords)

    def __eq__(self,other):
        return ( self.__class__ == other.__class__ ) and self._coords == other._coords

    def __ne__(self,other):
        return ( self.__class__ != other.__class__ ) or self._coords != other._coords

    def coords(self,dim=0):
        return self._coords



def ns(vector):   ## norm squared
    return vector.norm()**2



def cone_on_face(face,vertex):
    if face.dim == 0:
        return Polyhedron(1,[face,vertex])
    else:
        facets = [face]
        for f in face.facets(face.dim-1):
            facets.append(cone_on_face(f,vertex))
        return Polyhedron(face.dim+1,facets)

def double_cone_on_face(face,vertex):
    if face.dim == 0:
        return Polyhedron(1,[face,vertex])
    else:
        facets = [face]
        for f in face.facets(face.dim-1):
            facets.append(cone_on_face(f,vertex))
        return Polyhedron(face.dim+1,facets)

def quick_hull(points, dim):
    # choose dim+1 points for the initial hull. We might as well pick the ones of largest norm.
    v_max = sorted([(ns(v),v) for v in points],key=lambda x:x[0])[-1][1]
    B_points = [v_max]
    while len(B_points)<dim:
        A = Mat([v.transpose() for v in B_points]).transpose()
        Proj_P = A*(A.transpose()*A).inv()*A.transpose() #projection to the subspace generated by B_points
        perp_norms = dict([(v,ns(v-Proj_P*v)) for v in points])
        max_perp = sorted([(perp_norms[v],v) for v in points],key=lambda x:x[0])[-1][1]
        B_points.append(max_perp)
    #O = Vertex(IMat([0 for i in range(dim)]))
    F = simplex(B_points,dim-1)
    facets = []
    for set1,set2 in partitions(F.vertices()):
        facets.append(simplex([-v.coords() for v in set1]+[v.coords() for v in set2],dim-1))
        facets.append(simplex([-v.coords() for v in set2]+[v.coords() for v in set1],dim-1))
    hull = Polyhedron(dim,facets)
    above = {} # this dictionary will associate with each face the vertices that are above it 
                 # (i.e., those having positive inner product with the outward normal of the face)
    for F in hull.facets(dim-1):
        above[F] = outside_of(F,points) 
    points = set([p for F in hull.facets(hull.dim-1) for p in above[F]]) # we can get rid of points not above at least one face
    while len(points) > 0: # as long as there are still points outside of the hull, expand faces to points that lie above them
        for F in hull.facets(dim-1):
            if len(above[F]) > 0:
                p = above[F][-1] # pick the point furthest from F (i.e., the last in the list)
        hull,points,above = expand(hull,p,points,above)

    #### ***** still need to combine co-planar faces. Compute a normal for each face, group by scalar multiple, combine faces in group.********

    return hull

def expand(hull,point,points,above):
    visible = [face for face in hull.facets(hull.dim-1) if point in above[face]] # faces visible from point
    horizons = [] # facets of dimension dim-2 that are on the boundary of the union of visible faces as seen from point
    for face in visible:
        for subface in face.facets(hull.dim-2):
            on_faces = [f for f in visible if hull.is_subfacet(subface,f)]
            if len(on_faces) == 1: # only add to horizons if it is not shared by multiple faces (so it's not in the interior)
                horizons.append(subface)
    faces = hull.facets(hull.dim-1) # all faces of current hull
    for face in visible:
        faces.remove(face) #remove visible faces as they are now covered by the new faces we'll create below
        del above[face] # also remove these from the above dictionary
    points.discard(point)
    for h in horizons:
        new_face = cone_on_face(h,Vertex(point)) # new faces come from coning over the horizons to point
        above[new_face] = outside_of(new_face,points) # populate above[new_face] with points that lie above new_face
        faces.append(new_face) # add to faces each new face

    hull = Polyhedron(hull.dim,faces) # construct the new hull
    points = set([p for F in hull.facets(hull.dim-1) for p in above[F]])
    return hull, points, above


def simplex(points,dim):
    assert len(points) == dim+1
    if dim == 1:
        P = Polyhedron(1,[Vertex(points[0]),Vertex(points[1])])
        return P
    else:
        facets = [simplex(f,dim-1) for f in list(combinations(points,dim))]
        P = Polyhedron(dim,facets)
        return P

def outside_of(face,points):
    outside = []
    basepoint = face.vertices()[0]  # pick a vertex to translate everything to the origin by
    pre_basis = face.neighbors(basepoint)
    dim = face.dim
    basis = [v.coords()-basepoint.coords() for v in pre_basis][:dim]
    A = Mat([v.transpose() for v in basis]).transpose()
    Proj_P = A*((A.transpose()*A).inv())*A.transpose()
    #C = Mat([0 for i in range(len(basis))])
    #A_t=A.transpose()

    #solutions, param = A_t.gauss_jordan_solve(C)  # solve the system of equations, returns parametrization of solutions
    #param_one = { tau:1 for tau in param } #set parameters to 0 to get a solution which will hopefully be an integer solution.
    #normal = solutions.xreplace(param_one)
    #if normal.dot(basepoint.coords()) < 0:
    #    normal = -normal
    normal = basepoint.coords() - Proj_P*basepoint.coords()
    for p in points:
        dot = normal.dot(p-basepoint.coords())
        if dot > 0:
            outside.append((dot,p))
    outside.sort(key=lambda x: x[0])
    return [pair[1] for pair in outside]  

class vector():
    pass

def partitions(collection):
    collection = list(collection)
    if len(collection) == 1:
        return [[ collection,[] ]]

    parts = []
    first = collection[0]
    for [set1, set2] in partitions(collection[1:]):
        # insert `first` in each of the subpartition's subsets
        parts.append([set1,set2+[first]])
        parts.append([set1+[first],set2])
    return parts








